@@ -1,14 +1,13 @@
 #!/usr/bin/env python
-# vim:fileencoding=utf-8
 # License: GPLv3 Copyright: 2020, Kovid Goyal <kovid at kovidgoyal.net>
 
 from typing import (
-    Collection, Dict, Generator, List, NamedTuple, Optional, Sequence, Tuple,
+    Any, Collection, Dict, Generator, List, NamedTuple, Optional, Sequence, Tuple,
     Union
 )
 
 from kitty.borders import BorderColor
-from kitty.constants import Edges, WindowGeometry
+from kitty.types import Edges, WindowGeometry
 from kitty.typing import EdgeLiteral, WindowType
 from kitty.window_list import WindowGroup, WindowList
 
@@ -87,6 +86,7 @@ class Pair:
         for q in root.self_and_descendants():
             if q.one is self or q.two is self:
                 return q
+        return None
 
     def remove_windows(self, window_ids: Collection[int]) -> None:
         if isinstance(self.one, int) and self.one in window_ids:
@@ -324,7 +324,7 @@ class Pair:
                 return 'top', 'bottom'
             return 'bottom', 'top'
 
-        geometries = dict((group.id, group.geometry) for group in all_windows.groups if group.geometry)
+        geometries = {group.id: group.geometry for group in all_windows.groups if group.geometry}
 
         def extend(other: Union[int, 'Pair', None], edge: EdgeLiteral, which: EdgeLiteral) -> None:
             if not ans[which] and other:
@@ -389,6 +389,9 @@ class SplitsLayoutOpts(LayoutOpts):
     def __init__(self, data: Dict[str, str]):
         self.default_axis_is_horizontal = data.get('split_axis', 'horizontal') == 'horizontal'
 
+    def serialized(self) -> Dict[str, Any]:
+        return {'default_axis_is_horizontal': self.default_axis_is_horizontal}
+
 
 class Splits(Layout):
     name = 'splits'
@@ -447,13 +450,12 @@ class Splits(Layout):
     ) -> None:
         horizontal = self.default_axis_is_horizontal
         after = True
-        if location is not None:
-            if location == 'vsplit':
-                horizontal = True
-            elif location == 'hsplit':
-                horizontal = False
-            if location in ('before', 'first'):
-                after = False
+        if location == 'vsplit':
+            horizontal = True
+        elif location == 'hsplit':
+            horizontal = False
+        elif location in ('before', 'first'):
+            after = False
         aw = all_windows.active_window
         if aw is not None:
             ag = all_windows.active_group
@@ -461,6 +463,10 @@ class Splits(Layout):
             group_id = ag.id
             pair = self.pairs_root.pair_for_window(group_id)
             if pair is not None:
+                if location == 'split':
+                    wwidth = aw.geometry.right - aw.geometry.left
+                    wheight = aw.geometry.bottom - aw.geometry.top
+                    horizontal = wwidth >= wheight
                 target_group = all_windows.add_window(window, next_to=aw, before=not after)
                 pair.split_and_add(group_id, target_group.id, horizontal, after)
                 return
@@ -546,3 +552,22 @@ class Splits(Layout):
                     if swap:
                         pair.one, pair.two = pair.two, pair.one
                     return True
+        return None
+
+    def layout_state(self) -> Dict[str, Any]:
+
+        def add_pair(p: Pair) -> Dict[str, Any]:
+            ans: Dict[str, Any] = {}
+            ans['horizontal'] = p.horizontal
+            ans['bias'] = p.bias
+            if isinstance(p.one, Pair):
+                ans['one'] = add_pair(p.one)
+            elif p.one is not None:
+                ans['one'] = p.one
+            if isinstance(p.two, Pair):
+                ans['two'] = add_pair(p.two)
+            elif p.one is not None:
+                ans['two'] = p.two
+            return ans
+
+        return {'pairs': add_pair(self.pairs_root)}
